"""Client Object to connect to API and relevant Exceptions."""
# Import Built-Ins
import logging

# Import Third-Party

# Import Homebrew
from hitbtc.connector import HitBTCConnector

# Init Logging Facilities
log = logging.getLogger(__name__)


class CredentialsError(ValueError):
    pass


class HitBTC:
    """HitBTC Websocket API Client class.

    Programmed using the official API documentation as a reference.

    Documentation can be found here:
        https://api.hitbtc.com/?python#socket-api-reference

    """

    def __init__(self, key=None, secret=None, raw=None, stdout_only=False, silent=False, url=None,
                 **conn_ops):
        """
        Initialize the instance.

        :param key: API Public Key
        :param secret: API Secret Key
        :param raw: Bool, whether or not to unpack data or pass it as is
        :param stdout_only: Bool, passing True will turn off placing data on self.conn.q
        :param silent: Bool, passing True turns off print() arguments
        :param url: URL of the websocket API. Defaults to wss://api.hitbtc.com/api/2/ws
        :param conn_ops: Optional Kwargs to pass to the HitBTCConnector object
        """
        self.conn = HitBTCConnector(url, raw, stdout_only, silent, **conn_ops)
        self.key = key
        self.secret = secret

    @property
    def credentials_given(self):
        """Assert if credentials are complete."""
        return self.key and self.secret

    def start(self):
        """Start the websocket connection."""
        self.conn.start()

    def stop(self):
        """Stop the websocket connection."""
        self.conn.stop()

    def login(self, key=None, secret=None, basic=None, custom_nonce=None):
        """
        Login using the WSS API.

        Offical Endpoint Documentation:
            https://api.hitbtc.com/?python#socket-session-authentication
        """
        if not self.credentials_given and not (key and secret):
            raise CredentialsError("Must give API key and Secret to login to API!")
        else:
            self.conn.authenticate(key or self.key, secret or self.secret, basic, custom_nonce)

    def request_currencies(self, custom_id=None, **params):
        """
        Request currencies currently listed at HitBTC.

        Offical Endpoint Documentation:
            https://api.hitbtc.com/?python#get-currencies
        """
        self.conn.send('getCurrencies', custom_id, **params)

    def request_symbols(self, custom_id=None, **params):
        """
        Request symbols currently traded at HitBTC.

        Offical Endpoint Documentation:
            https://api.hitbtc.com/?python#get-symbols
        """
        self.conn.send('getSymbols', custom_id, **params)

    def request_trades(self, custom_id=None, **params):
        """
        Request trades executed at HitBTC.

        Offical Endpoint Documentation:
            https://api.hitbtc.com/?python#get-trades
        """
        self.conn.send('getTrades', custom_id=custom_id, **params)

    def request_balance(self, custom_id=None, **params):
        """
        Request your account's balance.

        This requires you to be logged-in! Call ``login()`` first.

        Offical Endpoint Documentation:
            https://api.hitbtc.com/?python#get-trading-balance
        """
        self.conn.send('getTradingBalance', custom_id=custom_id, **params)

    def request_active_orders(self, custom_id=None, **params):
        """
        Request your account's active orders.

        This requires you to be logged-in! Call ``login()`` first!

        Offical Endpoint Documentation:
            https://api.hitbtc.com/?python#get-active-orders-2
        """
        self.conn.send('getOrders', custom_id=custom_id, **params)

    def subscribe_reports(self, cancel=False, custom_id=None, **params):
        """
        Request a stream of your account's order activities.

        This requires you to be logged-in! Call ``HitBTC.login()`` first!

        Offical Endpoint Documentation:
            https://api.hitbtc.com/?python#subscribe-to-reports
        """
        method = 'subscribeReports'
        if cancel:
            method = 'un' + method
        self.conn.send(method, custom_id=custom_id, **params)

    def subscribe_ticker(self, cancel=False, custom_id=None, **params):
        """Request a stream for ticker data.

        Offical Endpoint Documentation:
            https://api.hitbtc.com/?python#subscribe-to-ticker
        """
        method = 'subscribeTicker'
        if cancel:
            method = 'un' + method
        self.conn.send(method, custom_id=custom_id, **params)

    def subscribe_book(self, cancel=False, custom_id=None, **params):
        """Request a stream for order book data.

        Offical Endpoint Documentation:
            https://api.hitbtc.com/?python#subscribe-to-orderbook
        """
        method = 'subscribeOrderbook'
        if cancel:
            method = 'un' + method
        self.conn.send(method, custom_id=custom_id, **params)

    def subscribe_trades(self, cancel=False, custom_id=None, **params):
        """Request a stream for trade data.

        Offical Endpoint Documentation:
            https://api.hitbtc.com/?python#subscribe-to-trades
        """
        method = 'subscribeTrades'
        if cancel:
            method = 'un' + method
        self.conn.send(method, custom_id=custom_id, **params)

    def subscribe_candles(self, cancel=False, custom_id=None, **params):
        """Request a stream for candle data.

        Offical Endpoint Documentation:
            https://api.hitbtc.com/?python#subscribe-to-candles
        """
        method = 'subscribeCandles'
        if cancel:
            method = 'un' + method
        self.conn.send(method, custom_id=custom_id, **params)

    def place_order(self, custom_id=None, **params):
        """
        Place a new order via Websocket.

        Offical Endpoint Documentation:
            https://api.hitbtc.com/?python#place-new-order
        """
        self.conn.send('newOrder', custom_id=custom_id, **params)

    def cancel_order(self, custom_id=None, **params):
        """
        Cancel an order via Websocket.

        Offical Endpoint Documentation:
            https://api.hitbtc.com/?python#cancel-order
        """
        self.conn.send('cancelOrder', custom_id=custom_id, **params)

    def replace_order(self, custom_id=None, **params):
        """
        Replace an existing order via Websocket.

        Offical Endpoint Documentation:
            https://api.hitbtc.com/?python#cancel-replace-orders
        """
        self.conn.send('cancelReplaceOrder', custom_id=custom_id, **params)



"""Connector Base class."""

# pylint: disable=too-many-arguments

# Import Built-Ins
import logging
from queue import Queue
from threading import Thread, Timer
import multiprocessing as mp

import json
import time
import ssl

# Import Third-Party
import websocket

# import Third-party
from pysher import Pusher

# Import home-grown

# Init Logging Facilities
log = logging.getLogger(__name__)


class WebSocketConnector:
    """Websocket Connection Thread.

    Inspired heavily by ekulyk's PythonPusherClient Connection Class
    https://github.com/ekulyk/PythonPusherClient/blob/master/pusherclient/connection.py

    Data received is available by calling WebSocketConnection.recv()
    """

    # pylint: disable=too-many-instance-attributes, too-many-arguments,unused-argument

    def __init__(self, url, timeout=None, q_maxsize=None, reconnect_interval=None, log_level=None):
        """Initialize a WebSocketConnector Instance.

        :param url: websocket address, defaults to v2 websocket.
        :param timeout: timeout for connection; defaults to 10s
        :param reconnect_interval: interval at which to try reconnecting;
                                   defaults to 10s.
        :param log_level: logging level for the connection Logger. Defaults to
                          logging.INFO.
        :param args: args for Thread.__init__()
        :param kwargs: kwargs for Thread.__ini__()
        """
        # Queue used to pass data up to Node
        self.q = Queue(maxsize=q_maxsize or 100)

        # Connection Settings
        self.url = url
        self.conn = None

        # Connection Handling Attributes
        self._is_connected = False
        self.disconnect_called = False
        self.reconnect_required = False
        self.reconnect_interval = reconnect_interval if reconnect_interval else 10
        self.paused = False

        # Setup Timer attributes
        # Tracks API Connection & Responses
        self.ping_timer = None
        self.ping_interval = 120

        # Set up history of sent commands for re-subscription
        self.history = []

        # Tracks Websocket Connection
        self.connection_timer = None
        self.connection_timeout = timeout if timeout else 10

        # Tracks responses from send_ping()
        self.pong_timer = None
        self.pong_received = False
        self.pong_timeout = 30

        self.log = logging.getLogger(self.__module__)
        self.log.setLevel(level=log_level if log_level else logging.INFO)
        if log_level == logging.DEBUG:
            websocket.enableTrace(True)

        formatter = logging.Formatter('%(asctime)s:%(name)s:%(levelname)s\t%(message)s')
        file_handler = logging.FileHandler(filename='wss.log', mode='w+')
        file_handler.setLevel(level=log_level if log_level else logging.DEBUG)
        file_handler.setFormatter(formatter)
        self.log.addHandler(file_handler)

    def stop(self):
        """Wrap around disconnect()."""
        self.disconnect()

    def disconnect(self):
        """Disconnect from the websocket connection and joins the Thread."""
        self.reconnect_required = False
        self.disconnect_called = True
        self._is_connected = False
        if self.conn:
            self.conn.close()

    def reconnect(self):
        """Issue a reconnection by setting the reconnect_required event."""
        # Reconnect attempt at self.reconnect_interval
        self.reconnect_required = True
        self._is_connected = False
        if self.conn:
            self.conn.close()

    def _connect(self):
        """Create a websocket connection.

        Automatically reconnects connection if it was severed unintentionally.
        """
        self.conn = websocket.WebSocketApp(
            self.url,
            on_open=self._on_open,
            on_message=self._on_message,
            on_error=self._on_error,
            on_close=self._on_close
        )

        ssl_defaults = ssl.get_default_verify_paths()
        sslopt_ca_certs = {'ca_certs': ssl_defaults.cafile}
        self.conn.run_forever(sslopt=sslopt_ca_certs)

        while self.reconnect_required:
            if not self.disconnect_called:
                self.log.info("Attempting to connect again in %s seconds.", self.reconnect_interval)
                time.sleep(self.reconnect_interval)

                # We need to set this flag since closing the socket will
                # set it to False
                self.conn.keep_running = True
                self.conn.run_forever(sslopt=sslopt_ca_certs)

    def run(self):
        """Run the main method of thread."""
        self._connect()

    def _on_message(self, ws, message):
        """Handle and pass received data to the appropriate handlers.

        Resets timers for time-out countdown and logs exceptions during parsing.

        All messages are time-stamped

        :param ws: Websocket obj
        :param message: received data as bytes
        :return:
        """
        self._stop_timer()

        raw, received_at = message, time.time()

        try:
            data = json.loads(raw)
        except json.JSONDecodeError as e:
            # Something wrong with this data, log and discard
            self.log.exception("Exception %s for data %s; Discarding..", e, raw)
            return
        finally:
            # We've received data, reset timers
            self._start_timer()
        self.pass_up(data, received_at)

    def _on_close(self, ws, *args):
        """Log the close and stop the time-out countdown.

        Execute when the connection is closed.

        :param ws: Websocket obj
        :param *args: additional arguments
        """
        self.log.info("Connection closed")
        self._stop_timer()

    def _on_open(self, ws):
        """Log connection status, set Events for _connect(), start timers and send a test ping.

        Execute on opening a new connection.

        If the connection was previously severed unintentionally, it re-subscribes
        to the channels by executing the commands found in self.history, in
        chronological order.

        :param ws: Webscoket obj
        """
        self.log.info("Connection opened")
        self._is_connected = True
        self._start_timer()
        if self.reconnect_required:
            self.log.info("Reconnection successful, re-subscribing to"
                          "channels..")
            for cmd in self.history:
                self.send(cmd)

    def _on_error(self, ws, error):
        """Log the error, reset the self._is_connected flag and issue a reconnect.

        Callback executed on connection errors.

        Issued by setting self.reconnect_required.

        :param ws: Websocket obj
        :param error: Error message
        """
        self.log.info("Connection Error - %s", error)
        self._is_connected = False
        self.reconnect_required = True

    def _stop_timer(self):
        """Stop connection timer."""

        if self.connection_timer:
            self.connection_timer.cancel()

    def _start_timer(self):
        """Reset and start timer for API connection."""
        self._stop_timer()

        # Automatically reconnect if we didnt receive data
        self.connection_timer = Timer(self.connection_timeout,
                                      self._connection_timed_out)
        self.connection_timer.start()

    def send(self, data):
        """Send the given Payload to the API via the websocket connection.

        Furthermore adds the sent payload to self.history.

        :param data: data to be sent
        :return:
        """
        if self._is_connected:
            payload = json.dumps(data)
            self.history.append(data)
            self.conn.send(payload)
        else:
            log.error("Cannot send payload! Connection not established!")

    def pass_up(self, data, recv_at):
        """Pass data up to the client via the internal Queue().

        :param data: data to be passed up
        :param recv_at: float, time of reception
        :return:
        """
        self.q.put(data, recv_at)

    def recv(self, block=True, timeout=None):
        """Wrap for self.q.get().

        :param block: Whether or not to make the call to this method block
        :param timeout: Value in seconds which determines a timeout for get()
        :return:
        """
        return self.q.get(block, timeout)

    def _connection_timed_out(self):
        """Issue a reconnection.

        :return:
        """
        self.reconnect()


class WebSocketConnectorThread(WebSocketConnector, Thread):
    """Thread-based WebsocketConnector."""

    def __init__(self, url, timeout=None, q_maxsize=None, reconnect_interval=None, log_level=None,
                 **kwargs):
        """Initialize the instance."""
        super(WebSocketConnectorThread, self).__init__(url, timeout=timeout, q_maxsize=q_maxsize,
                                                       reconnect_interval=reconnect_interval,
                                                       log_level=log_level)
        Thread.__init__(self, **kwargs)
        self.daemon = True

    def disconnect(self):
        """Disconnect from the websocket and join thread."""
        super(WebSocketConnectorThread, self).disconnect()
        Thread.join(self, timeout=1)


class WebSocketConnectorProcess(WebSocketConnector, mp.Process):
    """Process-based websocket connector."""

    def __init__(self, url, timeout=None, q_maxsize=None, reconnect_interval=None, log_level=None,
                 **kwargs):
        """Initialize the instance."""
        super(WebSocketConnectorProcess, self).__init__(url, timeout=timeout, q_maxsize=q_maxsize,
                                                        reconnect_interval=reconnect_interval,
                                                        log_level=log_level)
        mp.Process.__init__(self, **kwargs)
        self.daemon = True

    def disconnect(self):
        """Disconnect from the websocket and join the process."""
        super(WebSocketConnectorProcess, self).disconnect()
        mp.Process.join(self, timeout=1)



"""
Message templates to log when handling responses to requests that are SUCCESFUL.
Failed requests are logged using the error code contained in the response and its related message.
"""
resp_get_currency = '{currency}:\n' \
               '\t{fullname}({id}):' \
               '\tIs a cryptocurrency: {crypto}\n' \
               '\tDeposits available: {payinEnabled}\n' \
               '\tpayinPaymentId available: {payinPaymentId}\n' \
               '\tRequired confirmations on deposit: {payinConfirmations}\n' \
               '\tWithdrawals available: {payoutEnabled}\n' \
               '\tpayoutIsPaymentId available: {payoutIsPaymentId}\n' \
               '\tTransfers enabled: {transferEnabled}\n'

resp_get_currencies = '{fullname}({id}):' \
                 '\tIs a cryptocurrency: {crypto}\n' \
                 '\tDeposits available: {payinEnabled}\n' \
                 '\tpayinPaymentId available: {payinPaymentId}\n' \
                 '\tRequired confirmations on deposit: {payinConfirmations}\n' \
                 '\tWithdrawals available: {payoutEnabled}\n' \
                 '\tpayoutIsPaymentId available: {payoutIsPaymentId}\n' \
                 '\tTransfers enabled: {transferEnabled}\n'

resp_get_symbol = '{id}:\n' \
             '\tBase currency: {baseCurrency}\n' \
             '\tQuote currency: {quoteCurrency}\n' \
             '\tMinimum quantity increment: {quantityIncrement}\n' \
             '\tTick size: {tickSize}\n' \
             '\tMaker fee: {takeLiquidityRate}\n' \
             '\tTaker fee: {provideLiquidityRate}\n' \
             '\tFee currency: {feeCurrency}\n'

resp_get_symbols = '{id}:\n' \
              '\tBase currency: {baseCurrency}\n' \
              '\tQuote currency: {quoteCurrency}\n' \
              '\tMinimum quantity increment: {quantityIncrement}\n' \
              '\tTick size: {tickSize}\n' \
              '\tMaker fee: {takeLiquidityRate}\n' \
              '\tTaker fee: {provideLiquidityRate}\n' \
              '\tFee currency: {feeCurrency}\n'

resp_get_trades = 'Trade ID ({id}):' \
             '\tPrice: {price}\n' \
             '\tSize: {quantity}\n' \
             '\tSide: {side}\n' \
             '\tTimestamp: {timestamp}\n'

resp_get_active_orders = 'Trade ID ({id}):' \
                    '\tOrder type: {type}\n' \
                    '\tStatus: {status}\n' \
                    '\tPrice: {price}\n' \
                    '\tSize: {quantity}\n' \
                    '\tSide: {side}\n' \
                    '\tCumulative size: {cumQunatity}\n' \
                    '\tTime in Force: {timeInForce}\n' \
                    '\tCreated at: {timestamp}\n' \
                    '\tUpdated at: {timestamp}\n' \
                    '\tClient Order ID: {clientOrderId}\n' \
                    '\tOriginal Request Client Order ID: {originalRequestClientOrderId}\n' \
                    '\tReport type: {reportType}\n'

resp_get_trading_balance = 'Wallet: {}\n' \
                      '\tAvailable: {available}\n' \
                      '\tReserved: {reserved}\n'

resp_place_order = 'Successfully placed a new order via websocket!\n' \
              '\tTrade ID ({id}):' \
              '\t\tOrder type: {type}\n' \
              '\t\tStatus: {status}\n' \
              '\t\tPrice: {price}\n' \
              '\t\tSize: {quantity}\n' \
              '\t\tSide: {side}\n' \
              '\t\tCumulative size: {cumQunatity}\n' \
              '\t\tTime in Force: {timeInForce}\n' \
              '\t\tCreated at: {timestamp}\n' \
              '\t\tUpdated at: {timestamp}\n' \
              '\t\tClient Order ID: {clientOrderId}\n' \
              '\t\tReport type: {reportType}\n'

resp_cancel_order = 'Successfully cancelled an order via websocket!\n' \
              '\tTrade ID ({id}):' \
              '\t\tOrder type: {type}\n' \
              '\t\tStatus: {status}\n' \
              '\t\tPrice: {price}\n' \
              '\t\tSize: {quantity}\n' \
              '\t\tSide: {side}\n' \
              '\t\tCumulative size: {cumQunatity}\n' \
              '\t\tTime in Force: {timeInForce}\n' \
              '\t\tCreated at: {timestamp}\n' \
              '\t\tUpdated at: {timestamp}\n' \
              '\t\tClient Order ID: {clientOrderId}\n' \
              '\t\tReport type: {reportType}\n'

resp_cancel_replace_order = 'Successfully replaced an order via websocket!\n' \
                       '\tTrade ID ({id}):' \
                       '\t\tOrder type: {type}\n' \
                       '\t\tStatus: {status}\n' \
                       '\t\tPrice: {price}\n' \
                       '\t\tSize: {quantity}\n' \
                       '\t\tSide: {side}\n' \
                       '\t\tCumulative size: {cumQunatity}\n' \
                       '\t\tTime in Force: {timeInForce}\n' \
                       '\t\tCreated at: {timestamp}\n' \
                       '\t\tUpdated at: {timestamp}\n' \
                       '\t\tClient Order ID: {clientOrderId}\n' \
                       '\t\tOriginal Request Client Order ID: {originalRequestClientOrderId}\n' \
                       '\t\tReport type: {reportType}\n'

resp_subscribe_ticker = 'Succesfully subscribed to {symbol} ticker data!'
resp_subscribe_book = 'Succesfully subscribed to {symbol} order book data!'
resp_subscribe_trades = 'Succesfully subscribed to {symbol} trade data!'
resp_subscribe_candles = 'Succesfully subscribed to {symbol} candle data!'
resp_subscribe_reports = 'Succesfully subscribed to account reports!'

response_types = {'getCurrency': resp_get_currency, 'getCurrencies': resp_get_currencies,
                  'getSymbol': resp_get_symbol, 'getSymbols': resp_get_symbols,
                  'getTrades': resp_get_trades,
                  'getOrders': resp_get_active_orders,
                  'getTradingBalance': resp_get_trading_balance,
                  'subscribeTicker': resp_subscribe_ticker,
                  'subscribeOrderbook': resp_subscribe_book,
                  'subscribeTrades': resp_subscribe_trades,
                  'subscribeCandles': resp_subscribe_candles,
                  'subscribeReports': resp_subscribe_reports,
                  'newOrder': resp_place_order, 'cancelOrder': resp_cancel_order,
                  'cancelReplaceOrder': resp_cancel_replace_order}



"""HitBTC Connector which pre-formats incoming data to the CTS standard."""

import logging
import time
import json
import hmac
import hashlib
from threading import Timer
from collections import defaultdict

from hitbtc.wss import WebSocketConnectorThread
from hitbtc.utils import response_types

log = logging.getLogger(__name__)


class HitBTCConnector(WebSocketConnectorThread):
    """Class to pre-process HitBTC data, before putting it on the internal queue.

    Data on the queue is available as a 3-item-tuple by default:
        symbol, data_type, data

    Whereas data can either be the extracted ``params`` data from a streamed notification object, or
    a tuple of ``(request, response)``, where ``request`` is the original payload sent by the
    client and ``response`` the related response object from the server.

    You can disable extraction and handling by passing 'raw=True' on instantiation. Note that this
    will also turn off recording of sent requests, as well all logging activity.
    """

    def __init__(self, url=None, raw=None, stdout_only=False, silent=False, **conn_ops):
        """Initialize a HitBTCConnector instance."""
        url = url or 'wss://api.hitbtc.com/api/2/ws'
        super(HitBTCConnector, self).__init__(url, **conn_ops)
        self.books = defaultdict(dict)
        self.requests = {}
        self.raw = raw
        self.logged_in = False
        self.silent = silent
        self.stdout_only = stdout_only

    def put(self, item, block=False, timeout=None):
        """Place the given item on the internal q."""
        if not self.stdout_only:
            self.q.put(item, block, timeout)

    def echo(self, msg):
        """Print message to stdout if ``silent`` isn't True."""
        if not self.silent:
            print(msg)

    def _start_timers(self):
        """Reset and start timers for API connection."""
        self._stop_timers()

        # Automatically reconnect if we didnt receive data
        self.connection_timer = Timer(self.connection_timeout,
                                      self._connection_timed_out)
        self.connection_timer.start()

    def _stop_timers(self):
        """Stop connection timer."""
        if self.connection_timer:
            self.connection_timer.cancel()

    def _on_message(self, ws, message):
        """Handle and pass received data to the appropriate handlers."""

        if not self.raw:
            decoded_message = json.loads(message)
            if 'jsonrpc' in decoded_message:
                if 'result' in decoded_message or 'error' in decoded_message:
                    self._handle_response(decoded_message)
                else:
                    try:
                        method = decoded_message['method']
                        symbol = decoded_message['params'].pop('symbol')
                        params = decoded_message.pop('params')
                    except Exception as e:
                        self.log.exception(e)
                        self.log.error(decoded_message)
                        return
                    self._handle_stream(method, symbol, params)
        else:
            self.put(message)

    def _handle_response(self, response):
        """
        Handle JSONRPC response objects.

        Acts as a pre-sorting function and determines whether or not the response is an error
        message, or a response to a succesful request.
        """
        try:
            i_d = response['id']
        except KeyError as e:
            self.log.exception(e)
            self.log.error("An expected Response ID was not found in %s", response)
            raise

        try:
            request = self.requests.pop(i_d)
        except KeyError as e:
            log.exception(e)
            log.error("Could not find Request relating to Response object %s", response)
            raise

        if 'result' in response:
            self._handle_request_response(request, response)
        elif 'error' in response:
            self._handle_error(request, response)

    def _handle_request_response(self, request, response):
        """
        Handle responses to succesful requests.

        Logs messages and prints them to screen.

        Finally, we'll put the response and its corresponding request on the internal queue for
        retrieval by the client.
        """
        method = request['method']

        try:
            msg = response_types[method]
        except KeyError as e:
            log.exception(e)
            log.error("Response's method %s is unknown to the client! %s", method)
            return

        if method.startswith('subscribe'):
            if 'symbol' in response['params']:
                formatted_msg = msg.format({'symbol': response['params']['symbol']})
            else:
                formatted_msg = msg
            self.log.info(formatted_msg)
            self.echo(formatted_msg)
        else:
            text = "Sucessfully processed %s request:\n" % method
            if method.startswith('get'):
                # loop over item in response['result'] for:
                # getSymbols, getTrades, getTradingBalance, getOrders
                for item in response['result']:
                    text += msg.format(item)
                self.log.info(text)
                self.echo(text)
            else:
                # Format messages for these using response['result'] directly
                # (place, cancel, replace, getSymbol, getCurrency)
                text += msg.format(response['result'])
                self.log.info(text)
                self.echo(text)
        self.log.debug("Request: %r, Response: %r", request, response)
        self.put(('Response', 'Success', (request, response)))

    def _handle_error(self, request, response):
        """
        Handle Error messages.

        Logs the corresponding requests and the error code and error messages, and prints them to
        the screen.

        Finally, we'll put the response and its corresponding request on the internal queue for
        retrieval by the client.
        """
        err_message = "{code} - {message} - {description}!".format(response['error'])
        err_message += " Related Request: %r" % request
        self.log.error(err_message)
        self.echo(err_message)
        self.put(('Response', 'Failure', (request, response)))

    def _handle_stream(self, method, symbol, params):
        """Handle streamed data."""
        self.put((method, symbol, params))

    def send(self, method, custom_id=None, **params):
        """
        Send the given Payload to the API via the websocket connection.

        :param method: JSONRPC method to call
        :param custom_id: custom ID to identify response messages relating to this request
        :param kwargs: payload parameters as key=value pairs
        """
        payload = {'method': method, 'params': params, 'id': custom_id or int(10000 * time.time())}
        if not self.raw:
            self.requests[payload['id']] = payload
        self.log.debug("Sending: %s", payload)
        self.conn.send(json.dumps(payload))

    def authenticate(self, key, secret, basic=False, custom_nonce=None):
        """Login to the HitBTC Websocket API using the given public and secret API keys."""
        if basic:
            algo = 'BASIC'
            skey = secret
            payload = {'sKey': skey}
        else:
            algo = 'HS256'
            nonce = custom_nonce or str(round(time.time() * 1000))
            raw_sig = (key + nonce).encode(encoding='UTF-8')
            signature = hmac.new(secret, raw_sig, hashlib.sha256).hexdigest()
            payload = {'nonce': nonce, 'signature': signature}

        payload['algo'] = algo
        payload['pKey'] = key
        self.send('login', **payload)








from setuptools import setup

VERSION = '1.0.0'

setup(name='hitbtc',
      version=VERSION,
      description='HitBTC Websocket API Client',
      author='Nils Diefenbach',
      author_email='23okrs20+gitlab@mykolab.com',
      packages=['hitbtc'],
      classifiers=['Programming Language :: Python :: 3 :: Only'],
      install_requires=['websocket-client'],
      package_data={'': ['*.md', '*.rst']})

